@@ -4,25 +4,23 @@ Creation of the `Layout` instance for the Python input/REPL.
 from __future__ import unicode_literals
 
 from prompt_toolkit.enums import DEFAULT_BUFFER, SEARCH_BUFFER
-from prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, Always, HasFocus, Condition
+from prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, HasFocus, Condition
 from prompt_toolkit.key_binding.vi_state import InputMode
 from prompt_toolkit.layout.containers import Window, HSplit, VSplit, FloatContainer, Float, ConditionalContainer, ScrollOffsets
 from prompt_toolkit.layout.controls import BufferControl, TokenListControl, FillControl
 from prompt_toolkit.layout.dimension import LayoutDimension
 from prompt_toolkit.layout.lexers import SimpleLexer
-from prompt_toolkit.layout.margins import Margin
+from prompt_toolkit.layout.margins import PromptMargin
 from prompt_toolkit.layout.menus import CompletionsMenu, MultiColumnCompletionsMenu
-from prompt_toolkit.layout.highlighters import SearchHighlighter, SelectionHighlighter, MatchingBracketHighlighter, ConditionalHighlighter
-from prompt_toolkit.layout.processors import HighlightMatchingBracketProcessor, ConditionalProcessor, AppendAutoSuggestion
+from prompt_toolkit.layout.processors import ConditionalProcessor, AppendAutoSuggestion, HighlightSearchProcessor, HighlightSelectionProcessor, HighlightMatchingBracketProcessor
 from prompt_toolkit.layout.screen import Char
 from prompt_toolkit.layout.toolbars import CompletionsToolbar, ArgToolbar, SearchToolbar, ValidationToolbar, SystemToolbar, TokenListToolbar
 from prompt_toolkit.layout.utils import token_list_width
-from prompt_toolkit.mouse_events import MouseEventTypes
 from prompt_toolkit.reactive import Integer
 from prompt_toolkit.selection import SelectionType
-from prompt_toolkit.utils import get_cwidth
 
-from ptpython.filters import HasSignature, ShowSidebar, ShowSignature, ShowDocstring
+from .filters import HasSignature, ShowSidebar, ShowSignature, ShowDocstring
+from .utils import if_mousedown
 
 from pygments.lexers import PythonLexer
 from pygments.token import Token
@@ -71,18 +69,31 @@ def python_sidebar(python_input):
                 (T, '\n'),
             ])
 
-        def append(selected, label, status):
+        def append(index, label, status):
+            selected = index == python_input.selected_option_index
+
+            @if_mousedown
+            def select_item(cli, mouse_event):
+                python_input.selected_option_index = index
+
+            @if_mousedown
+            def goto_next(cli, mouse_event):
+                " Select item and go to next value. "
+                python_input.selected_option_index = index
+                option = python_input.selected_option
+                option.activate_next()
+
             token = T.Selected if selected else T
 
             tokens.append((T, ' >' if selected else '  '))
-            tokens.append((token.Label, '%-24s' % label))
-            tokens.append((token.Status, ' '))
-            tokens.append((token.Status, '%s' % status))
+            tokens.append((token.Label, '%-24s' % label, select_item))
+            tokens.append((token.Status, ' ', select_item))
+            tokens.append((token.Status, '%s' % status, goto_next))
 
             if selected:
                 tokens.append((Token.SetCursorPosition, ''))
 
-            tokens.append((token.Status, ' ' * (14 - len(status))))
+            tokens.append((token.Status, ' ' * (13 - len(status)), goto_next))
             tokens.append((T, '<' if selected else ''))
             tokens.append((T, '\n'))
 
@@ -91,8 +102,7 @@ def python_sidebar(python_input):
             append_category(category)
 
             for option in category.options:
-                append(i == python_input.selected_option_index,
-                       option.title, '%s' % option.get_current_value())
+                append(i, option.title, '%s' % option.get_current_value())
                 i += 1
 
         tokens.pop()  # Remove last newline.
@@ -126,7 +136,7 @@ def python_sidebar_navigation(python_input):
 
         # Show navigation info.
         tokens.extend([
-            (T.Separator , ' ' * 43 + '\n'),
+            (T.Separator, ' ' * 43 + '\n'),
             (T, '    '),
             (T.Key, '[Arrows]'),
             (T, ' '),
@@ -238,7 +248,7 @@ def signature_toolbar(python_input):
             ~IsDone())
 
 
-class PromptMargin(Margin):
+class PythonPromptMargin(PromptMargin):
     """
     Create margin that displays the prompt.
     It shows something like "In [1]:".
@@ -246,34 +256,17 @@ class PromptMargin(Margin):
     def __init__(self, python_input):
         self.python_input = python_input
 
-    def _get_prompt_style(self):
-        return self.python_input.all_prompt_styles[self.python_input.prompt_style]
+        def get_prompt_style():
+            return python_input.all_prompt_styles[python_input.prompt_style]
 
-    def get_width(self, cli):
-        # Take the width from the first line.
-        text = ''.join(t[1] for t in self.python_input.get_input_prompt_tokens(cli))
-        return get_cwidth(text)
+        def get_prompt(cli):
+            return get_prompt_style().in_tokens(cli)
 
-    def create_margin(self, cli, window_render_info, width, height):
-        style = self._get_prompt_style()
+        def get_continuation_prompt(cli, width):
+            return get_prompt_style().in2_tokens(cli, width)
 
-        # First line.
-        tokens = style.in_tokens(cli)
-
-        # Next lines. (Show line numbering when numbering is enabled.)
-        tokens2 = style.in2_tokens(cli, width)
-        show_numbers = self.python_input.show_line_numbers
-        visible_line_to_input_line = window_render_info.visible_line_to_input_line
-
-        for y in range(1, min(window_render_info.content_height, height)):
-            tokens.append((Token, '\n'))
-            if show_numbers:
-                line_number = visible_line_to_input_line.get(y) or 0
-                tokens.append((Token.LineNumber, ('%i ' % (line_number + 1)).rjust(width)))
-            else:
-                tokens.extend(tokens2)
-
-        return tokens
+        super(PythonPromptMargin, self).__init__(get_prompt, get_continuation_prompt,
+                show_numbers=Condition(lambda cli: python_input.show_line_numbers))
 
 
 def status_bar(key_bindings_manager, python_input):
@@ -282,11 +275,13 @@ def status_bar(key_bindings_manager, python_input):
     """
     TB = Token.Toolbar.Status
 
+    @if_mousedown
     def toggle_paste_mode(cli, mouse_event):
-        if mouse_event.event_type == MouseEventTypes.MOUSE_DOWN:
-            python_input.paste_mode = not python_input.paste_mode
-        else:
-            return NotImplemented
+        python_input.paste_mode = not python_input.paste_mode
+
+    @if_mousedown
+    def enter_history(cli, mouse_event):
+        python_input.enter_history(cli)
 
     def get_tokens(cli):
         python_buffer = cli.buffers[DEFAULT_BUFFER]
@@ -310,8 +305,8 @@ def status_bar(key_bindings_manager, python_input):
             append((TB, '[Ctrl-W] Cut [Meta-W] Copy [Ctrl-Y] Paste [Ctrl-G] Cancel'))
         else:
             result.extend([
-                (TB.Key, '[F3]'),
-                (TB, ' History '),
+                (TB.Key, '[F3]', enter_history),
+                (TB, ' History ', enter_history),
                 (TB.Key, '[F6]', toggle_paste_mode),
                 (TB, ' ', toggle_paste_mode),
             ])
@@ -338,15 +333,13 @@ def get_inputmode_tokens(cli, python_input):
 
     :param cli: `CommandLineInterface` instance.
     """
+    @if_mousedown
     def toggle_vi_mode(cli, mouse_event):
-        if mouse_event.event_type == MouseEventTypes.MOUSE_DOWN:
-            python_input.vi_mode = not python_input.vi_mode
-        else:
-            return NotImplemented
+        python_input.vi_mode = not python_input.vi_mode
 
     token = Token.Toolbar.Status
 
-    mode = python_input.key_bindings_manager.get_vi_state(cli).input_mode
+    mode = cli.vi_state.input_mode
     result = []
     append = result.append
 
@@ -381,12 +374,10 @@ def show_sidebar_button_info(python_input):
     Create `Layout` for the information in the right-bottom corner.
     (The right part of the status bar.)
     """
+    @if_mousedown
     def toggle_sidebar(cli, mouse_event):
         " Click handler for the menu. "
-        if mouse_event.event_type == MouseEventTypes.MOUSE_DOWN:
-            python_input.show_sidebar = not python_input.show_sidebar
-        else:
-            return NotImplemented
+        python_input.show_sidebar = not python_input.show_sidebar
 
     token = Token.Toolbar.Status
 
@@ -486,29 +477,27 @@ def create_layout(python_input, key_bindings_manager,
             BufferControl(
                 buffer_name=DEFAULT_BUFFER,
                 lexer=lexer,
-                highlighters=[
+                input_processors=[
+                    ConditionalProcessor(
+                        processor=HighlightSearchProcessor(preview_search=True),
+                        filter=HasFocus(SEARCH_BUFFER),
+                    ),
+                    HighlightSelectionProcessor(),
                     # Show matching parentheses, but only while editing.
-                    ConditionalHighlighter(
-                        highlighter=MatchingBracketHighlighter(chars='[](){}'),
+                    ConditionalProcessor(
+                        processor=HighlightMatchingBracketProcessor(chars='[](){}'),
                         filter=HasFocus(DEFAULT_BUFFER) & ~IsDone() &
                             Condition(lambda cli: python_input.highlight_matching_parenthesis)),
-                    ConditionalHighlighter(
-                        highlighter=SearchHighlighter(preview_search=Always()),
-                        filter=HasFocus(SEARCH_BUFFER)),
-                    SelectionHighlighter(),
-                ],
-                input_processors=[
                     ConditionalProcessor(
                         processor=AppendAutoSuggestion(),
                         filter=~IsDone())
                 ] + extra_buffer_processors,
                 menu_position=menu_position,
-                wrap_lines=Condition(lambda cli: python_input.wrap_lines),
 
                 # Make sure that we always see the result of an reverse-i-search:
-                preview_search=Always(),
+                preview_search=True,
             ),
-            left_margins=[PromptMargin(python_input)],
+            left_margins=[PythonPromptMargin(python_input)],
             # Scroll offsets. The 1 at the bottom is important to make sure the
             # cursor is never below the "Press [Meta+Enter]" message which is a float.
             scroll_offsets=ScrollOffsets(bottom=1, left=4, right=4),
@@ -516,6 +505,7 @@ def create_layout(python_input, key_bindings_manager,
             get_height=(lambda cli: (
                 None if cli.is_done or python_input.show_exit_confirmation
                         else input_buffer_height)),
+            wrap_lines=Condition(lambda cli: python_input.wrap_lines),
         )
 
     return HSplit([
@@ -544,7 +534,8 @@ def create_layout(python_input, key_bindings_manager,
                               bottom=1,
                               content=exit_confirmation(python_input)),
                         Float(bottom=0, right=0, height=1,
-                              content=meta_enter_message(python_input)),
+                              content=meta_enter_message(python_input),
+                              hide_when_covering_content=True),
                         Float(bottom=1, left=1, right=0, content=python_sidebar_help(python_input)),
                     ]),
                 ArgToolbar(),
